# import numpy as np 
import autograd.numpy as np
import autograd.numpy.random as npr
from autograd.scipy.stats import multivariate_normal

def test_flows():
	from vi_families import RealNVP
	q = RealNVP(5, 16, 2, 0.1, 3)

	ϕ = q.initial_params()

	z = q.sample(ϕ, (4,2))
	lq = q.log_prob(ϕ, z)
	ε, _ = q.inverse_transform(ϕ, z)
	ζ, _ = q.forward_transform(ϕ, ε)

	results = {
		"z" : z,
		"ζ" : ζ,
		"ε" : ε,
		"lq" : lq,
		"δ" : np.linalg.norm((ζ - z))/np.linalg.norm(z),
	}

	[print(f"{k}: {v}") for k,v in results.items()];

def test_gaussian():

	from vi_families import Gaussian

	q = Gaussian(3)

	ϕ = q.initial_params()

	z = q.sample(ϕ, (4,2))
	lq = q.log_prob(ϕ, z)

	lg = multivariate_normal.logpdf(z, np.zeros(3), np.eye(3))
	# lg = g.pdf(z)

	results = {
		"z" : z,
		"lq" : lq,
		"lg" : lg,
		"δ" : np.linalg.norm((lq - lg))/np.linalg.norm(lq),
	}
	[print(f"{k}: {v}") for k,v in results.items()];

def test_diagonal():

	from vi_families import Diagonal

	q = Diagonal(3)

	ϕ = q.initial_params()
	z = q.sample(ϕ, (4,2))
	lq = q.log_prob(ϕ, z)

	lg = multivariate_normal.logpdf(z, np.zeros(3), np.eye(3))
	# lg = g.pdf(z)

	results = {
		"δ" : np.linalg.norm((lq - lg))/np.linalg.norm(lq),
		"δ_abs" : np.linalg.norm((lq - lg)),
	}
	[print(f"{k}: {v}") for k,v in results.items()];


# test_diagonal()



import pystan
import warnings
import pickle
import os
import collections
import scipy
import warnings
import traceback
import functools
import autograd
import autograd.numpy as np
import vistan.utilities as utils

###########################################################################
#  stuff related to loading stan models and data
###########################################################################

def is_good_model(code, data, model_name = None, verbose = True):
    if model_name is not None:

        model_name = model_name.replace("-","_")
    # TODO : make sure data is deterministic
    try:

        model = Model(code, data, model_name, verbose = verbose)
        # print(model.zlen)
        print(model.keys)
        exit()
        Z_nuts = model.sampling(iter=20, verbose = verbose)
        Z_advi = model.advi(iter=20, verbose = verbose)
        # Z_mf   = model.mf(iter=20)
        # print(Z_nuts)
        # print(Z_nuts.shape)
        # print(Z_mf)
        # print(Z_mf.shape)
        for Z in [Z_advi]:
            if Z.shape[1]!=Z_nuts.shape[1]:
                raise ValueError("NUTS and VI not same dimensions")

        # also check that optimization works
        z,rez  = model.argmax(True,method='BFGS',gtol=1e-3,maxiter=20) 

        if not rez.success:
            pass
            # print(rez)
            # raise ValueError("optimization failed")
        if len(z) != Z_nuts.shape[1]:
            print("some")
            raise ValueError("NUTS and argmax not same dimensions")

        return True

    except:
        print(f"Error during model check...")
        # print(traceback.print_exc())
        raise
        # return False

###########################################################################
#  stuff related to compiling and caching stan models
###########################################################################

# removes whitespace and such to prevent extra compiling
# this is only used for hashing (original code is compiled)


# import Cython
# get_model_name(model_name, code):
#     model_code_stripped = standardize_code(model_code)
#     code_hash = hashlib.md5(model_code_stripped.encode('ascii')).hexdigest()
#     if model_name is None:
#         return model_name+"-"+
#     else:
#         cache_fn = 'data/cached-models/{}-{}.pkl'.format(model_name, code_hash)

def get_compiled_model(model_code, model_name=None, verbose = False, **kwargs):
    """Use just as you would `stan` from pystan"""
    if not os.path.exists('data/cached-models'):
        print('creating cached-models/ to save compiled stan models')
        os.makedirs('data/cached-models')

    cache_fn = f'data/cached-models/{utils.get_cache_fname(model_name, model_code)}.pkl'

    if os.path.isfile(cache_fn):
        with open(cache_fn, 'rb') as f:
            sm = pickle.load(f)
    else:
        try:
            print("Cached model not found. Recompiling...")            
            with utils.suppress_stdout_stderr(verbose = verbose):
                sm = pystan.StanModel(model_code=model_code, model_name=model_name, **kwargs)
        except:
            print("Error during compilation...")
            print('Could not compile code for ' + model_code, ". Trying turning verbose for better debugging.")
            raise
        print("Caching model...")
        with open(cache_fn, 'wb') as f:
            pickle.dump(sm, f)        
    return sm

###########################################################################
#  here's the object provided to the user-- a logp function that hooks into autograd
###########################################################################

class Model:
    def __init__(self, model_code, data, model_name=None, verbose = True, debugging_verbose = False):
        extra_compile_args = ['-O1','-w','-Wno-deprecated'] # THIS IS SLOW! TURN OPTIMIZATION ON SOMEDAY!

        self.data = data
        self.model_name = model_name
        self.model_code = model_code

        self.sm = get_compiled_model(model_code=self.model_code, extra_compile_args=extra_compile_args,\
                model_name=self.model_name, verbose = verbose)
        try:
            with utils.suppress_stdout_stderr(verbose = debugging_verbose):
                self.fit = self.sm.sampling(data=self.data, iter=100, chains=1, init=0)
        except:
            print('Could not init model for ' + model_code)
            print('Try turning debugging_verbose on for better debugging.')
            raise 

        self.keys = self.fit.unconstrained_param_names()
        self.zlen = len(self.keys)

        # with utils.suppress_stdout_stderr(verbose = debugging_verbose):
        #     results = self.sm.vb(data = self.data, algorithm = "meanfield", iter = 10)
        #     self.advi_param_names = results['sampler_param_names']
        # except:
        #     print('Could not run advi on model ' + model_code)
        #     print('Try turning debugging_verbose on for better debugging.')
        #     raise 

    def get_constrained_param_shapes(self):
        results = self.fit.extract()
        N = self.get_n_samples(results)
        constrained_param_shapes = collections.OrderedDict()

        for k,v in results.items():
            if k == "lp__":
                continue 
            assert(N == v.shape[0])
            constrained_param_shapes[k] = v.shape[1:]
        return constrained_param_shapes


    def z0(self):
        """
            Get a random initialization for the latent parameters.  Useful for MAP inference.
        """
        return np.random.randn(self.zlen)

    def sampling(self, verbose = True, **kwargs):
        """
            A function to access PyStan MCMC sampling techniques. 
            Returns samples in the unconstrained domain, in the form an array.
            See array_to_dict function to convert into dictionary format.  
        """
        try: 
            assert kwargs.get('iter',2) >=2
            warnings.simplefilter('ignore')
            with utils.suppress_stdout_stderr(verbose):
                self.fit = self.sm.sampling(data=self.data,**kwargs)
            # print(self.fit.extract().keys())
            rez = self.unconstrain(self.fit.extract())
            warnings.simplefilter('default')
        except:
            raise
        return rez

    def argmax(self,with_rez=False,method='BFGS',**kwargs):
        """
            A function to obtain MAP estimate using scipy.optimize.minimize.
            Arguments:
                with_rez :  Boolean
                            If True, returns the entire results object along with z_map.
                method :    String
                            One of suggested_solvers = ['CG','BFGS','L-BFGS-B','TNC','SLSQP'].
                kwargs:     arguments for scipy.optimize.minimize. see https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
                            for more details 
            Returns:
                z_map:      parameter value at Maximum-a-posteriori
                rez:        result object from scipy.optimize.minimize
        """
        suggested_solvers = ['CG','BFGS','L-BFGS-B','TNC','SLSQP']
        if not method in suggested_solvers:
            warnings.warn("Solver" + str(method) +
            " passed to argmax not in suggested list" + str(suggested_solvers)+
            '\nsee list at: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html')
        z = self.z0()
        obj = autograd.value_and_grad(lambda z : -self.logp(z),0)
        kwargs['disp'] = False
        rez = scipy.optimize.minimize(obj,z,method=method,jac=True,options=kwargs)
        z = rez.x
        if with_rez:
            return rez.x, rez
        else:
            return rez.x

        
    def advi(self, algorithm='fullrank', verbose = True, **kwargs):
        """
            A function to run Stan's variational Bayes method (ADVI)
            Arguments:
                algorithm:  "fullrank" or "meanfield"
                verbsoe:    Boolean
                            If True, prints the optimization messages from ADVI
                kwargs:     arguments for vb see https://pystan.readthedocs.io/en/latest/api.html#pystan.StanModel.vb
                            for more details 
            Returns:
                samples:    samples from the final posterior
        """
        try:
            with utils.suppress_stdout_stderr(verbose = verbose):
                rez = self.sm.vb(data=self.data,algorithm=algorithm,**kwargs)
        except:
            print('ADVI failed with error...\n', traceback.print_exc())
            return np.array([[]])
        # def pystan_vb_extract(results):
        samples = np.array(rez["sampler_params"])[:-1,:].T
        # samples = rez['sampler_params']
        # param_specs = rez['sampler_param_names']
        # params1 = self.method1(samples, param_specs)
        # params2 = self.method2(samples)
        # p1 = self.unconstrain(params1)
        # p2 = self.unconstrain(params2)        #returning unconstrain for internal consistency
        # return p1, p2 
        return self.unconstrain(self.method2(samples))

    def method1(self, samples, param_specs):

        N = len(samples[0])


        # first pass, calculate the shape
        param_shapes = collections.OrderedDict()
        for param_spec in param_specs:
            splt = param_spec.split('[')
            name = splt[0]
            if len(splt) > 1:
                idxs = [int(i) for i in splt[1][:-1].split(',')]  # no +1 for shape calculation because pystan already returns 1-based indexes for vb!
            else:
                idxs = ()
            param_shapes[name] = np.maximum(idxs, param_shapes.get(name, idxs))

        # create arrays
        params = collections.OrderedDict([(name, np.nan * np.empty((N, ) +\
                                 tuple(shape))) for name, shape in param_shapes.items()])

        # second pass, set arrays
        for param_spec, param_samples in zip(param_specs, samples):
            splt = param_spec.split('[')
            name = splt[0]
            if len(splt) > 1:
                idxs = [int(i) - 1 for i in splt[1][:-1].split(',')]  # -1 because pystan returns 1-based indexes for vb!
            else:
                idxs = ()
            params[name][(..., ) + tuple(idxs)] = param_samples

        return params

    def constrained_array_to_dict(self, samples):
        assert samples.ndim == 2
        N = samples.shape[0]

        constrained_param_shapes = self.get_constrained_param_shapes()
        params = collections.OrderedDict()
        idx = 0
        for param_name, param_shape in constrained_param_shapes.items():
            nparam = int(np.prod(param_shape))
            params[param_name] = np.reshape(samples[:, idx:idx+nparam], (N, *param_shape), order = "F")
            idx+=nparam
        assert idx == samples.shape[1]
        return params

    def mf(self,**kwargs):
        """
            A function to run Stan's variational Bayes method (ADVI) -- meanfield
        """
        return self.advi(algorithm='meanfield', verbose = True, **kwargs)


    def constrain(self, z):
        """
            A function to constrain the parameters from unconstrained to original constrained range
            as defined in the model.
            Arguments:
                z:  samples in array np.ndarray format
            Returns:
                samples: unconstrained samples np.ndarray format
        """

        assert z.ndim == 2
        # assert z.shape[-1] == self.zlen
        return np.array([self.fit.constrain_pars(np.asarray(z_,order='C')) for z_ in z])

    def get_n_samples(self, z):
        """
            A function to get the number of samples. 
            Arguments:
                z:  samples in array np.ndarray or dictionary format
            Returns:
                N: number of samples
        """

        if not isinstance(z, dict):
            assert z.ndim == 2
            return z.shape[0]
        else:
            return list(z.values())[0].shape[0]

    # def array_to_dict(self, z):
    #     """
    #         A function to convert parameters from np.ndarray format to dictionary. 
    #         Keys are inferred from the model definition.
    #         Arguments:
    #             z:  samples in array np.ndarray format
    #         Returns:
    #             samples: samples in dictionary format
    #     """

    #     assert self.get_n_samples(z) >= 1 
    #     assert z.shape[-1] == self.zlen

    #     return collections.OrderedDict({k:z[:, i] for i,k in enumerate(self.keys)})

    # def dict_to_array(self, z):
    #     """
    #         A function to convert parameters from dictionary format to np.ndarray. 
    #         Keys are used from the dictionary.
    #         Arguments:
    #             z: samples in dictionary format
    #         Returns:
    #             samples:  samples in array np.ndarray format
    #     """
    #     if not isinstance(z, dict):
    #         raise ValueError
    #     print(z.keys())
    #     for k in z.keys():
    #         print(z[k].shape)
    #     s = np.array([v  for k.v in z.items()])
    #     assert s.shape[-1] == self.zlen
    #     return s

    def unconstrain(self, z):
        """
            A function to unconstrain the parameters from original constrained range to unconstrained
            domain. Uses the model definition from Stan to transform.
            Arguments:
                z:  samples in array np.ndarray or dictionary format
            Returns:
                samples: unconstrained samples np.ndarray format
        """
        assert isinstance(z, dict)
        N = self.get_n_samples(z)
        # print(N)
        # if not isinstance(z, (dict,)):
        #     z = self.array_to_dict(z)
        # else:
            # not required. Stan takes care of this internally.
            # if "lp___" in list(z.keys()):
            #     del z['lp__']

        return np.array([self.fit.unconstrain_pars({k:v[n] for k,v in z.items()}) \
                                for n in range(N)])


    def log_prob(self, z):
        """
            A simple function to get batched sample evaluation for models.
            Arguments:
                z:  unconstrained samples in array np.ndarray 
            Returns:
                log p: model log p(z,x) evaluation at unconstrained z samples. Adjust transform in used under the hood. 
        """
        orig_samples_shape = z.shape
        assert(orig_samples_shape[-1] == self.zlen)
        lp = np.array([self.logp(z_) for z_ in z.reshape(-1, self.zlen)])

        return lp.reshape(orig_samples_shape[:-1]) 

    @autograd.primitive
    def logp(self,z):
        assert(z.ndim==1)
        assert(len(z)==self.zlen)
        try: # try to evaluate logp with stan. if you fail, just return nan
            return self.fit.log_prob(z,adjust_transform=True)
        except:
            return np.nan
        return rez
        
    def glogp(self,z):
        assert(len(z)==self.zlen)
        rez_from_stan = self.fit.grad_log_prob(z,adjust_transform=True)
        return rez_from_stan.reshape(z.shape)

    def vjpmaker(argnum,rez,stuff,args):
        obj = stuff[0]
        z   = stuff[1]
        if np.isnan(rez):
            return lambda gg : 0*z # special gradient for nan case
        else:
            return lambda gg : gg*Model.glogp(obj,z)

###########################################################################
#  magic that makes autograd treat stan model logp as primitive function
###########################################################################

autograd.extend.defvjp_argnum(Model.logp, Model.vjpmaker)




import argparse
import time
import hjson 
import traceback
from utilities import open_pickled_files

def run_vi_instance(hyper_params):

    import autograd.numpy.random as npr
    
    import interface
    from inference import  vi_on_stan_model

    print("Trying to load the model...")

    model = interface.get_logp([hyper_params['model_root_dir'], hyper_params['model_name']])

    print("Model successfully imported to autograd...")

    npr.seed(hyper_params['seed'])    #set up the seed and call the inference function 

    print("Starting Variational Inference...")

    vi_on_stan_model(model = model, hyper_params = hyper_params)

def run_vi(hyper_params):

    from utilities.result_helper import save_current_dict

    for exp in hyper_params['step_size_exp_range']:
    
        hyper_params['step_size_exp'] = exp
    
        try :
            print("Starting optimization...")

            start = time.time()

            run_vi_instance(hyper_params = hyper_params.copy())

            print("The time taken for the process : ",time.time() - start)
    
        except Exception:
    
            print(f"Error during the training of model {hyper_params['model_name']}...")

            print(traceback.print_exc())
    
    print(f"Optimization for {hyper_params['model_name']} done....")
    if "test" not in hyper_params['uniq_name']:
        save_current_dict(hyper_params = hyper_params)



if __name__ == "__main__" :
    

    parser = argparse.ArgumentParser()

    parser.add_argument('-uname', '--uname', type=str, required=True, 
        help="Unique name used for saving and retrieving.") 
    parser.add_argument('-ids', '--model_ids', type=int, required=False,
        help="Id of the models to experiment with. Should use Ids from table 1 in the paper.", nargs = "*") 
    parser.add_argument('-root', '--root_dir', type=str, required=False, 
        default = "./models",
        help="Root directory that contains the Stan code and the data for the model.") 
    parser.add_argument('-name', '--model_name', type=str, required=False, 
        help="Name of the model in the root directory. The same name is used to identify the Stan code and data files.") 
    parser.add_argument('-pconfig', '--pconfig', type=str, required=False, 
        default= '../ezstan/config.hjson', help="path to the .json config file")

    args = parser.parse_args()
    
    if args.pconfig:
        with open(args.pconfig, 'r') as f:
            hyper_params = hjson.load(f)
    else:
        raise NotImplementedError

    hyper_params['uniq_name'] = args.uname

    hyper_params['model_ids'] = args.model_ids


    
    if hyper_params['model_ids'] is None:
    
        assert args.model_name is not None
        assert args.root_dir is not None

        hyper_params['model_id'] = None
        hyper_params['model_root_dir'] = args.root_dir
        hyper_params['model_name'] = args.model_name

        run_vi(hyper_params)

    else: 

        model_orig_id_dict = (open_pickled_files("../data/utilities/good_model_new_id_to_old_id_dict.pkl"))
        model_names = (open_pickled_files("../data/utilities/good_model_name.pkl"))

        for m in hyper_params['model_ids']:

            hyper_params['model_id'] = model_orig_id_dict[m]
            hyper_params['model_root_dir'] = './models'
            hyper_params['model_name'] = model_names[hyper_params['model_id']]

            run_vi(hyper_params)



import os
import pickle
import time
import functools
import operator
import collections
import traceback 
import hashlib
import re
import scipy
import autograd 
import autograd.numpy as np 
import autograd.numpy.random as npr
import autograd.misc.optimizers as optim

########################################################################################
######################### Generic Inference Utilities ###############################
########################################################################################
def update_hparams(model, hparams):
    hparams['latent_dim'] = model.zlen

    update_hparams_method(hparams)

    hparams['num_copies_training'] = (hparams['per_iter_sample_budget']//
                                hparams['M_training'])
    # If advi is not used, then we follow the the step-size scaling 
    # scheme of Agrawal et al. 2020 

    hparams['step_size'] = hparams['step_size_base']/\
                                (hparams['step_size_scale']**hparams['step_size_exp'])

    hparams['step_size'] = hparams['step_size']/hparams['latent_dim']

def update_hparams_method(hparams):
    if hparams['method'] == 'advi':
        hparams.update({
                'advi_use' : True,
                'advi_adapt_step_size' : True,
                'advi_adapt_step_size_range' : [100, 10, 1, 0.1, 0.01],
                'advi_step_size' : 1,
                'vi_family' : "gaussian",
                'grad_estimator' : "closed-form-entropy",
                'optimizer' : 'advi',
                'advi_convergence_threshold' : 0.001,
                'advi_adapt_step_size_num_iters' : 100,
                'M_training' : 1,
                'LI' : False,
                'per_iter_sample_budget':100,

            })

    elif hparams['method'] == 'gaussian':
        hparams.update({
                'advi_use' : False,
                'vi_family' : "gaussian",
                'grad_estimator' : "DReG",
                'optimizer' : 'adam',
                'M_training' : 10,
                'LI' : False,
                'step_size_exp' : 2,
                'per_iter_sample_budget':100,

            })

    elif hparams['method'] == 'meanfield':
        hparams.update({
                'advi_use' : False,
                'vi_family' : "diagonal",
                'grad_estimator' : "DReG",
                'optimizer' : 'adam',
                'M_training' : 1,
                'LI' : False,
                'step_size_exp' : 2,
                'per_iter_sample_budget':100,

            })

    elif hparams['method'] == 'flows':
        hparams.update({
                'advi_use' : False,
                'vi_family' : "rnvp",
                'grad_estimator' : "DReG",
                'optimizer' : 'adam',
                'M_training' : 2,
                'step_size_exp' : 2,
                'LI' : False,
                'per_iter_sample_budget':100,

            })


    elif hparams['method'] == 'custom':
        pass

    else:
        raise NotImplementedError(f"Method not unsupported. Expected one\
                 of ['advi', 'flows', 'gaussian', 'meanfield', 'custom'], \
                 but got {hparams['method']}")



S = np.log(np.exp(1) - 1)

def good_iter(i):
    a = 10 ** np.floor(np.log10(i*1.0))
    return (i%a)==0

def log_add_exp(x1, x2):
    return np.maximum(x1,x2) + np.log1p(np.exp(-np.abs(x1-x2)))

def log_sub_exp(x1, x2):
    assert (all(x1>x2))
    return x1 + np.log1p(-np.exp(x2-x1))

def pos_diag(x):
    return log_add_exp(x, 0)

def pos_tril(x):
    return np.tril(x, -1) + np.diag(pos_diag(np.diag(x)))

def inv_pos_diag(x):
    return log_sub_exp(np.diag(x), 0)

def inv_pos_tril(x):
    return np.tril(x, -1) + np.diag(inv_pos_diag(x)) 

def mul_iterable(x):
    return functools.reduce(operator.mul, x, 1)
    
########################################################################################
######################### Laplace's Initialization Utilities ###########################
########################################################################################

def Hessian_finite_differences(z, grad_f, ε):

    D = len(z)
    H = np.zeros((D,D))

    for d in range(D):
        z_pos = z*1.0
        z_pos[d] += ε
        z_neg = z*1.0
        z_neg[d] -= ε

        H[:,d] = (grad_f(z_pos) - grad_f(z_neg))/(2*ε)

    return H

def get_laplaces_init_params(log_p, z_len, num_epochs, ε = 1e-4):        

    # Initialize to Laplace's method 
    # Conduct MAP inference using BFGS method
    # Set mu = z_MAP
    # Use finite difference to calculate Hessian matrix
    # Set L = cholesky(inv(-Hessian(mu)))

    z_0 = npr.rand(z_len)
    val_and_grad = autograd.value_and_grad(lambda z: -log_p(z)) # using minimize to maximize

    rez = scipy.optimize.minimize(val_and_grad, z_0, \
                    method='BFGS', jac=True,\
                    options={'maxiter':num_epochs, 'disp':True})

    mu = rez.x
    H = Hessian_finite_differences(z = mu, grad_f = autograd.grad(lambda z : log_p(z)),\
                                    ε = ε)

    try :
        neg_H_inv = np.linalg.inv(-H)
        L = np.linalg.cholesky(neg_H_inv) # -H_inv = inv(-H)
        L = inv_pos_tril(L) # modification to adjust for pos_tril function 

    except:
        print ('Using noisy unit covariance...')
        L = np.tril(0.1*npr.randn(z_len, z_len), -1) + np.eye(z_len)
        L = inv_pos_tril(L) # modification to adjust for pos_tril function 

    return mu, L

def standardize_code(text):
    # text = re.sub('//.*?\n|/\*.*?\*/', '', text, flags=re.S)
    text = re.sub(r'//.*?\n|/\*.*?\*/', '', text, flags=re.S)
    pat = re.compile(r'\s+')
    text = pat.sub('', text)
    return text


def get_cache_fname(model_name, model_code):
    model_code_stripped = standardize_code(model_code)
    code_hash = hashlib.md5(model_code_stripped.encode('ascii')).hexdigest()
    if model_name is None:
        return code_hash
    else:
        return model_name + "_"+ code_hash

def get_laplaces_init(log_p, z_len, num_epochs, ε, model_name, model_code):
    if check_laplaces_init_saved(model_name, model_code):
        return load_saved_laplaces_init(model_name, model_code)
    else:
        LI_params = get_laplaces_init_params(log_p, z_len, num_epochs, ε)
        save_laplaces_init(LI_params, model_name, model_code)
        return LI_params

def laplaces_init_dir():
    return 'data/laplaces_init/'

def laplaces_init_file(dir_name, model_name, model_code):
    return dir_name + get_cache_fname(model_name, model_code) +"_LI_params.pkl"

def check_laplaces_init_saved(model_name, model_code):
    """
    Check if stored Laplaces Initialization parameters  are available.
    """    
    file_name = laplaces_init_file(laplaces_init_dir(), model_name, model_code)
    if os.path.exists(file_name):
        return True
    return False

def load_saved_laplaces_init(model_name, model_code):
    file_name = laplaces_init_file(laplaces_init_dir(), model_name, model_code)
    if not os.path.exists(file_name):
        raise ValueError
    return open_pickled_files(file_name)

    

def save_laplaces_init(params,  model_name, model_code):  
    dir_name = laplaces_init_dir()
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    dump_pickled_files(filename = laplaces_init_file(dir_name, model_name, model_code),
        objects = params)


########################################################################################
######################### Optimization Utilities             ###########################
########################################################################################



def advi_asserts(hparams):

    assert(hparams['advi_use'] == 1)
    assert(hparams['vi_family'] == 'gaussian')
    assert(hparams['M_training'] == 1)
    assert(hparams['LI'] == 0)
    assert(hparams['grad_estimator'] == 'closed-form-entropy')
    assert(hparams['optimizer'] == 'advi')


def get_callback_arg_dict(hparams):

    if hparams['advi_use'] == True:

        buffer_len = np.int(max(0.01*hparams['max_iters']/hparams['advi_callback_iteration'] , 2))
        delta_results = collections.deque(maxlen = buffer_len)

        return {"delta_results" : delta_results, "hparams" : hparams}

    else:

        return {}

def run_optimization(objective_grad, init_params, step_size,\
             num_epochs, callback, optimizer):

    optimized_params = optimizer(grad = objective_grad, 
                                x0 = init_params, 
                                step_size = step_size,
                                num_iters = num_epochs, 
                                callback = callback)

    return optimized_params

def get_adapted_step_size(objective_grad, eval_function, init_params, \
                        optimizer, num_epochs, hparams):

    """
    Implements the adaptive step-size scheme from the PyStan version of ADVI.
    """

    init_elbo = eval_function(params = init_params)

    best_elbo = -1.0*np.inf

    best_step_size = 0

    print("############################################ ")
    print(f"########## Initial elbo: {init_elbo}")
    print("############################################ ")

    for i, step_size in enumerate(hparams['advi_adapt_step_size_range']):

        results = []

        print("############################################ ")
        print(f"########## Checking the step_size: {step_size}")
        print("############################################ ")

        try: 
            optimized_params = run_optimization(objective_grad = objective_grad, 
                                                init_params = init_params, 
                                                step_size = step_size, 
                                                num_epochs = hparams['advi_adapt_step_size_num_iters'], 
                                                callback = None, 
                                                optimizer = advi_optimizer)

        except Exception:
            print(f"Error occured during the optimization with step-size {step_size}...")
            print(traceback.print_exc())
            print(f"Using initial parameters instead for {step_size}...")
            optimized_params = init_params

        candidate_elbo  = eval_function(optimized_params)

        if np.isnan(candidate_elbo):

            candidate_elbo = -1.0*np.inf 

        if  (candidate_elbo < best_elbo) & \
            (best_elbo > init_elbo):

            assert(best_step_size!= 0)

            print("Best step_size found, best step_size : ", best_step_size)
            print("Best step_size found, best elbo : ", best_elbo)

            return best_step_size

        else:

            if  ((i+1) < len(hparams['advi_adapt_step_size_range'])):

                best_elbo = candidate_elbo
                best_step_size = step_size

            else:

                if candidate_elbo > init_elbo:

                    print("Best step_size found, best step_size : ", best_step_size)
                    print("Best step_size found, best elbo : ", best_elbo)

                    return best_step_size

                else :

                    print("ELBO value diverged for all step_sizes. Update step_size range")
                    exit()




def optimization_handler(objective_grad, eval_function, init_params, optimizer,\
                            num_epochs, step_size, callback, hparams, **kwargs):

    if  (hparams['advi_adapt_step_size']) & (hparams['advi_use']):

        step_size = get_adapted_step_size(objective_grad = objective_grad, 
                                            eval_function = eval_function,
                                            init_params = init_params,
                                            optimizer = optimizer, 
                                            num_epochs = num_epochs, 
                                            hparams = hparams)

    results = []
    t0 = time.time()
    optimized_params = run_optimization(objective_grad = objective_grad,
                                        init_params = init_params,
                                        num_epochs = num_epochs,
                                        step_size = step_size,
                                        callback = functools.partial(callback, 
                                                                    results = results,
                                                                    **get_callback_arg_dict(hparams)),
                                        optimizer = optimizer)

    tn = time.time() - t0
    return results, optimized_params

# def checkpoint(params, model, hparams, results, t0, n):

#     if hparams['check_point_use']==1:

#         if (n+1) in hparams['check_point_num_epochs']:

#             tn = time.time() - t0

#             save_results_parameters(hparams = hparams,
#                                     params = params,
#                                     model = model,
#                                     uniq_name = hparams['uniq_name']+"_check_point_"+str(n+1),
#                                     results = results,
#                                     time = tn/(n+1))

def callback(params, t, g, results, model, eval_function):

    results.append(eval_function(params))

    if good_iter(t+1):

        if np.isnan(results[-1]): 

            print("exiting optimization because nan encountered.")

            raise ValueError

        print("Iteration {} IWELBO (RUNNING AVERAGE) {}".format(t+1, np.mean(results)))
        print("Iteration {} IWELBO (CURRENT ESTIMATE) {}".format(t+1, results[-1]))

    # checkpoint(params, model, hparams, results, t0 = t0, n = t)

def relative_difference(curr, prev):

    return np.abs((curr-prev)/prev)

def advi_callback(params, t, g, results, delta_results, model, \
                            eval_function, hparams):

    results.append(eval_function(params))

    if (t+1)%hparams['advi_callback_iteration']==0:

        print(f"Iteration {t+1} log likelihood IWELBO (RUNNING AVERAGE) {np.nanmean(results)}")
        print(f"Iteration {t+1} log likelihood IWELBO (CURRENT AVERAGE) {(results[-1])}")

        if len(results) > hparams['advi_callback_iteration']:

            previous_elbo = results[-(hparams['advi_callback_iteration']+1)] 

        else: 

            previous_elbo = 0.0 

        current_elbo = results[-1]

        delta_results.append(relative_difference(previous_elbo, current_elbo))

        delta_elbo_mean = np.nanmean(delta_results)

        delta_elbo_median = np.nanmedian(delta_results)

        print(f"Iteration {t+1} Δ mean {delta_elbo_mean}")
        print(f"Iteration {t+1} Δ median {delta_elbo_median}")


        if  (   (delta_elbo_median <= hparams['advi_convergence_threshold'])|
                (delta_elbo_mean <= hparams['advi_convergence_threshold'])
            ):

            print("Converged according to ADVI metrics for Median/Mean")
            # break
            # tn = time.time() - t0

            # save_results_parameters (hparams = hparams,
            #                         params = params,
            #                         model = model,
            #                         uniq_name = hparams['uniq_name'] + str("_delta_convergence_"),
            #                         results = results,
            #                         time = tn/(t+1))
            # exit()
            return True
    return False

    # checkpoint(params, model, hparams, results, t0 = t0, n = t)

def advi_optimizer(grad, x0, callback, num_iters, step_size,\
                                 epsilon = 1e-16, tau = 1, alpha = 0.1):

    x, unflatten = autograd.misc.flatten(x0)

    s = np.zeros(len(x))

    for i in range(num_iters):

        g = autograd.misc.flatten(grad(unflatten(x), i))[0]

        if callback: 
            converged = callback(unflatten(x), i, unflatten(g))
            if converged :
                return unflatten(x)

        if i==0:

            s = g**2

        else:

            s = alpha*(g**2) + (1-alpha)*s

        x = x -  (step_size / np.sqrt(i+1.))*g/ (tau + np.sqrt(s))

    return unflatten(x)

def get_step_size(hparams):

    if hparams['advi_use'] == True:

        return hparams['advi_step_size']

    return hparams['step_size']


def get_optimizer(hparams):


    if hparams['optimizer']=="adam":
      
        return optim.adam

    elif hparams['optimizer']=="advi":

        return advi_optimizer

    else: 
        raise NotImplementedError


def get_callback(hparams):

    if hparams['advi_use'] == 0:
    
        return callback
    
    else: 
    
        return advi_callback


########################################################################################
######################### Generic Flow Utilities #######################################
########################################################################################


def coupling_layer_specifications(num_hidden_units, num_hidden_layers, z_len):
    """
    We specify the FNN based networks over here. A single network produce both s and t parts.
    Coupling Layer currently comprises of 2 transforms.
    """

    d_1 = np.int(z_len//2) 
    d_2 = np.int(z_len - d_1)
    
    coupling_layer_sizes = []
    coupling_layer_sizes.append([d_1] + num_hidden_layers*[num_hidden_units] + [2*d_2])
    coupling_layer_sizes.append([d_2] + num_hidden_layers*[num_hidden_units] + [2*d_1])
    return coupling_layer_sizes


########################################################################################
######################### Generic NN Utilities #########################################
########################################################################################

def relu(x):       
    return np.maximum(0, x)

def leakyrelu(x, slope = 0.01):       
    return np.maximum(0, x) + slope*np.minimum(0,x)

def tanh(x): 
    return np.tanh(x)

def softmax_matrix(x):
    assert x.ndim <=2
    y = x.T
    z = y - np.max(y, 0)
    z = np.exp(z)
    return (z/np.sum(z, 0)).T


########################################################################################
######################### Generic file Utilities #######################################
########################################################################################


def dump_pickled_files(filename, objects, protocol = None):
    with open(filename,"wb") as f:
        if protocol is None:
            pickle.dump(objects,f)
        else:
            pickle.dump(objects,f, protocol = protocol)

def open_pickled_files(filename, protocol = None):
    with open(filename,"rb") as f:
        if protocol is None:
            objects = pickle.load(f)
        else:
            objects = pickle.load(f, protocol = protocol)
        return objects


########################################################################################
######################### Generic Print Utilities #######################################
########################################################################################

def print_dict(dic):
    
    # hparams['step_size'] = hparams['step_size']/hparams['latent_dim']

    # using a fixed sample budget--we need to adjust the no. of training samples based on M 

    [print(f"{k} : {i}") for k, i in dic.items()];


class suppress_stdout_stderr(object):
    '''
    A context manager for doing a "deep suppression" of stdout and stderr in 
    Python, i.e. will suppress all print, even if the print originates in a 
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).      

    '''
    def __init__(self, verbose):
        self.verbose = verbose
        # Open a pair of null files
        self.null_fds =  [os.open(os.devnull,os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):

        if not self.verbose:

        # Assign the null pointers to stdout and stderr.
            os.dup2(self.null_fds[0],1)
            os.dup2(self.null_fds[1],2)

    def __exit__(self, *_):
        
        # Re-assign the real stdout/stderr back to (1) and (2)
        if not self.verbose:

            os.dup2(self.save_fds[0],1)
            os.dup2(self.save_fds[1],2)
            # Close all file descriptors
            for fd in self.null_fds + self.save_fds:
                os.close(fd)



import autograd.numpy as np
import autograd.numpy.random as npr
import functools 
import collections

import vistan.utilities as utils

class Dist():

    def __init__(self, zlen):
        
        self.zlen = zlen

    def initial_params(self):
        
        raise NotImplementedError

    def transform_params(self, params):
        
        raise NotImplementedError

    def get_params(self, params):

        return self.transform_params(params)

    def sample(self, params, sample_shape):

        raise NotImplementedError

    def log_prob(self, params, samples):

         raise NotImplementedError


S = np.log(np.exp(1)-1)


class Gaussian(Dist):

    def initial_params(self):
        
        return [np.zeros(self.zlen,).astype(float), S*np.eye(self.zlen).astype(float)]

    def transform_params(self, params):

        return params[0], utils.pos_tril(params[1])


    def sample(self, params, sample_shape = ()):

        if isinstance(sample_shape, int) : 
            sample_shape = (sample_shape,)

        mu, sig = self.get_params(params)
        shape = sample_shape+(self.zlen,)
        ε  = npr.randn(*shape)
        samples = mu + np.dot(ε, sig.T)
        
        return samples

    def log_prob(self, params, samples):

        mu, sig = self.get_params(params)

        Λ = np.linalg.inv(sig).T

        M = (samples-mu)

        a = self.zlen*np.log(2*np.pi)

        b = 2*np.sum(np.log((np.diag(sig))))

        c = np.sum(np.matmul(M,np.matmul(Λ, Λ.T)) * M, -1)

        return -0.5*(a+b+c) 

    def entropy(self, params):

        mu, sig = self.get_params(params)

        a = self.zlen*np.log(2*np.pi) + self.zlen

        b = 2*np.sum(np.log((np.diag(sig))))

        return 0.5*(a+b)



class Diagonal(Gaussian):

    def initial_params(self):
        
        return [np.zeros(self.zlen,).astype(float), S*np.ones(self.zlen,).astype(float)]

    def transform_params(self, params):
        
        return params[0], np.diag(utils.pos_diag(params[1]))


class Flows(Dist):

    def __init__(self, zlen):

        self.zlen = zlen
        self.base_dist = Gaussian(self.zlen)
        self.base_dist_params = self.base_dist.initial_params()

    def transform_params(self, params):

        return params

    def forward_transform(self, params, z):

        raise NotImplementedError

    def inverse_transform(self, params, x):

        raise NotImplementedError

    def sample(self, params, sample_shape):
        

        params = self.get_params(params)
        
        z_o = self.base_dist.sample(self.base_dist_params, sample_shape)

        samples, neg_log_det_J = self.forward_transform(params, z_o)

        return samples

    def log_prob(self, params, samples):

        params = self.get_params(params)

        z_o, neg_log_det_J = self.inverse_transform(params, samples)

        lq = self.base_dist.log_prob(self.base_dist_params, z_o)

        return lq + neg_log_det_J

def BinaryFlip(z, j):

    return np.concatenate(reversed(BinarySplit(z, j)), -1)

def ReverseBinaryFlip(z, j):

    return np.concatenate(reversed(ReverseBinarySplit(z, j)), -1)
    
def BinarySplit(z, j):

    D = z.shape[-1]

    d = D//2
    
    if D%2 == 1: d+=(np.int(j)%2)

    return np.array_split(z, [d], -1)

def ReverseBinarySplit(z, j):

    return BinarySplit(z, np.int(j)+1)

class RealNVP(Flows):

    def __init__(   self, 
                    num_transformations, 
                    num_hidden_units, 
                    num_hidden_layers, 
                    params_init_scale, 
                    zlen):
        
        self.num_transformations = num_transformations
        self.num_hidden_layers = num_hidden_layers
        self.num_hidden_units = num_hidden_units
        self.params_init_scale = params_init_scale

        super(RealNVP, self).__init__(zlen = zlen)
    
    def initial_params(self):

        def generate_net_st():

            coupling_layer_sizes = utils.coupling_layer_specifications(self.num_hidden_units, self.num_hidden_layers, self.zlen)

            init_st_params = []

            for layer_sizes in coupling_layer_sizes:

                 init_st_params.append([(self.params_init_scale * npr.randn(m, n),   # weight matrix
                                               self.params_init_scale * npr.randn(n))      # bias vector
                                              for m, n in zip(layer_sizes[:-1], layer_sizes[1:])])
            return init_st_params

        st = [generate_net_st() for i in range(self.num_transformations)]

        return st

    def apply_net_st(self, params, inputs):

        inpW, inpb = params[0]

        inputs = utils.leakyrelu(np.dot(inputs, inpW) + inpb)

        for W, b in params[1:-1]:
            outputs = np.dot(inputs, W) + b
            inputs = utils.leakyrelu(outputs)

        outW, outb = params[-1]
        outputs = np.dot(inputs, outW) + outb

        assert(outputs.shape[:-1] == inputs.shape[:-1])
        assert(outputs.shape[-1]%2 == 0)

        s,t = np.array_split(outputs, 2 , -1)

        assert(s.shape == t.shape)

        return utils.tanh(s),t

    def forward_transform(self, params, z):

        neg_log_det_J = np.zeros(z.shape[:-1])

        st_list = params

        for i in range(self.num_transformations):
            for j in range(2):

                z_1, z_2 = BinarySplit(z,j)

                s, t = self.apply_net_st( params = st_list[i][j], inputs = z_1)
                z = np.concatenate([z_1, z_2*np.exp(s) + t ], axis = -1)
                neg_log_det_J -= np.sum(s,axis=-1)  #Immediately update neg log jacobian with s outputs

                z = BinaryFlip(z, j)

        assert(z.shape[:-1] == neg_log_det_J.shape)
        return z, neg_log_det_J

    def inverse_transform(self, params, x):

        neg_log_det_J = np.zeros(x.shape[:-1])

        st_list = params
        
        for i in reversed(range(self.num_transformations)):
            for j in reversed(range(2)):

                x = ReverseBinaryFlip(x, j)

                x_1, x_2 = BinarySplit(x, j)

                s, t = self.apply_net_st( params = st_list[i][j], inputs = x_1)#output should be of the shape of x_1

                x = np.concatenate([x_1, (x_2 - t)*np.exp(-s) ], axis = -1)
                neg_log_det_J -= np.sum(s,axis=-1)

        assert(x.shape[:-1] == neg_log_det_J.shape)
        
        return x, neg_log_det_J


def get_var_dist(hyper_params):

    if hyper_params['vi_family'] == "gaussian":

        var_dist = Gaussian(hyper_params['latent_dim'])

    elif hyper_params['vi_family'] == "diagonal":

        var_dist = Diagonal(hyper_params['latent_dim'])

    elif hyper_params['vi_family'] == "rnvp":

        var_dist = RealNVP(num_transformations = hyper_params['rnvp_num_transformations'], 
                            num_hidden_units = hyper_params['rnvp_num_hidden_units'], 
                            num_hidden_layers = hyper_params['rnvp_num_hidden_layers'], 
                            params_init_scale = hyper_params['rnvp_params_init_scale'], 
                            zlen = hyper_params['latent_dim'])     
    else: 
        raise NotImplementedError
    
    return var_dist


class Posterior():

    def __init__(self, M_sampling, model, params, var_dist):

        self.M_sampling = M_sampling
        self.params = params
        self.model = model
        self.var_dist = var_dist


        self.log_p = self.model.log_prob
        self.log_q = self.var_dist.log_prob
        self.sample_q = self.var_dist.sample
        self.zlen = self.model.zlen


    def sample(self, num_samples, params = None, M_sampling = None, return_constrained = True):
        """
            A function to allows sampling from the posterior.

            Arguments : 

                num_samples:            # of samples 

                params:                 if None, then the stored optimized variational parameters
                                        are used to sample. Alternatively, one can provide 
                                        compatible parameters.

                M_sampling:             If None, then M_sampling is set to M_training the 
                                        variational inference was optimized with. Otherwise, a
                                        user can specify a different M_sampling. This scales the 
                                        sampling cost linearly.

                return_constrained:     If True, returns the constrained parameters in the same
                                        dictionary template as PyStan's .extract() method (without 'lp__'.) 
                                        If False, then samples are returned unconstrained in np.ndarray
                                        format such that z.shape = (num_samples, self.zlen) 
            Returns:

                z :                     If return_constrained is True, then a dictionary is the same template
                                        as PyStan's StanModelFit4.extract() method. 
                                        Else, returns the unconstrained samples in the np.ndarray format
                                        such that z.shape = (num_samples, self.zlen)

        """
        if params is None: 
            params = self.params

        if M_sampling is None : 
            M_sampling = self.M_sampling

        if  M_sampling== 1:
            samples =  self.sample_q(params, num_samples)

        else: 
            shape = (num_samples, M_sampling)
            samples = self.sample_q(params, shape)

            lp = self.log_p(samples)
            lq = self.log_q(params, samples)
            lR = lp - lq

            final_samples = []
            for i in range(lR.shape[0]):
                j = np.argmax(npr.multinomial(1, utils.softmax_matrix(lR[i])))
                final_samples.append(samples[i,j,:])

            samples =  np.array(final_samples).reshape(num_samples, self.zlen)

        if return_constrained == True:
            return self.constrained_array_to_dict(self.constrain(samples))
        else :
            return samples

    def log_prob(self, samples, params = None):

        if params is None: 
            params = self.params


        if self.M_sampling == 1:

            # log q is only valid if M_sampling = 1
            # if M_sampling > 1, then the overall q is different from the 
            # base distribution. See Divide and Couple, NeurIPS'19. 

            if isinstance(samples, dict):
                samples = self.unconstrain(samples)
            return self.log_q(params, samples)

        else :
            raise NotImplementedError

    def constrain(self, z):
        return self.model.constrain(z)

    def constrained_array_to_dict(self, z):
        assert z.ndim == 2
        return self.model.constrained_array_to_dict(z)

    def unconstrain(self, samples):
        assert isinstance(samples, dict)
        return self.model.unconstrain(samples)

def get_posterior(model, var_dist, params, M_sampling):

    q = Posterior(M_sampling = M_sampling, 
                    model = model, 
                    var_dist = var_dist, 
                    params = params,)
    return q



import os
import shutil
import traceback
from utilities import open_pickled_files    


def transfer_model_files_from_orig_to_model():
    # + Go through the models
    # + for each model
    #   + load the code, and the data
    #   + save the code and the data

    source_dir = "./eg_stan/"
    target_dir = "./models"
        
    file_names = os.listdir(source_dir)
        

    # shutil.move(source, dest)  

    good_model_paths = (open_pickled_files("../data/utilities/good_model_paths.pkl"))
    good_model_orig_id_dict = (open_pickled_files("../data/utilities/good_model_new_id_to_old_id_dict.pkl"))

    for m in range(1, 31):

        model_path = good_model_paths[good_model_orig_id_dict[m]]
        root, model_name = model_path
        source_dir = "./"+root
        file_names = os.listdir(source_dir)
        for file_name in file_names:

            try:

                shutil.move(os.path.join(source_dir, file_name), target_dir)

            except Exception:

                 print(traceback.print_exc())

def change_the_name_dict():
    good_model_names = (open_pickled_files("../data/utilities/good_model_name.pkl"))
    good_model_orig_id_dict = (open_pickled_files("../data/utilities/good_model_new_id_to_old_id_dict.pkl"))

    print(good_model_orig_id_dict)
def test_vistan():
    model = """
        data {
            int<lower=0> N; 
            int<lower=0,upper=1> switc[N];
        }
        parameters {
             real<lower=0,upper=1> beta1;
             real<lower=2,upper=2.4> beta2;
        } 
        model {
            switc ~ bernoulli(beta1);
        }
    """

    data = {"N" : 2, "switc": [1,0]}



# change_the_name_dict()
def test_pystan():
    import pystan
    import pickle
    import warnings
    from mystan import suppress_stdout_stderr, is_good_model
    from collections import OrderedDict
    model = """
        data {
            int<lower=0> N; 
            int<lower=0,upper=1> switc[N];
        }
        parameters {
             real<lower=0,upper=1> beta1;
             real<lower=2,upper=2.4> beta2;
        } 
        model {
            switc ~ bernoulli(beta1);
        }
    """

    data = {"N" : 2, "switc": [1,0]}

    print(is_good_model(model, data, model_name = "test_model"))
    exit()
    # def get_lop(model, data):

    try:
        with open("./test-model.pkl", "rb") as f:
            m = pickle.load(f)
    except:
        print("Did not find cached model. Re-compiling...")
        try: 
            with suppress_stdout_stderr(verbose = False):
                m = pystan.StanModel(model_code = model)
        except:
            print("Errors during recompile...")
            raise 
        print("Finished recompiling...")
        with open("./test-model.pkl", "wb") as f:
            pickle.dump(m, f)
        print("Model recompiled and cached...")

    with suppress_stdout_stderr(verbose = False):
        fit4 = m.sampling(data = data, chains = 1, iter = 10, seed = 0)
    # try : 
    #     with suppress_stdout_stderr(True):
    #         # print("blah blah")
    #         a = 1/0
    # except :
    #     pass
    # raise 
    # print("somehtign")
    # print (traceback.print_exc())
    # exit()

    print(fit4.extract()["beta1"].shape)   
    print(fit4.extract())   

    # exit()
    keys = fit4.unconstrained_param_names()
    z_len = len(keys)
    # print(z_len)
    # exit()
    """
    
    We want to constrain these parameters and then provide them in the shape of a dictionary
    
    """
    def constrain_params(z):

        assert z.shape[-1] == z_len

        return np.array([fit4.constrain_pars(np.asarray(z_, order = "C")) for z_ in z])

    def get_n_samples(z):

        assert z.ndim == 2
        return z.shape[0]

    def array_to_dict(z):

        assert z.shape[-1] == z_len
        assert get_n_samples(z) >= 1 

        dz = OrderedDict()

        for i,k in enumerate(keys):
            dz[k] = z[:,i] 

        return dz

    def unconstrain_params(z):

        N = get_n_samples(z)
        dz = array_to_dict(z)

        return np.array([fit4.unconstrain_pars({k:v[n] for k,v in dz.items()}) \
                                for n in range(N)])


    N = 5

    z = np.random.randn(N*2).reshape(N,2).astype(float)

    z_cons = (constrain_params(z))
    print(array_to_dict(z_cons))

    z_uncons = unconstrain_params(z_cons)
    print(array_to_dict(z_uncons))

    print(z)
    with suppress_stdout_stderr(verbose = False):
        rez = m.vb(data=data,algorithm="meanfield", output_samples = 10, sample_file = "./samples")
    z_vb = np.array(rez['sampler_params']).T[:,:-1]
    print(f"Unconstrained VB samples {z_vb}")
    print(f"Constrained VB samples {constrain_params(z_vb)}")
    print(f"Constrained VB samples {array_to_dict(z_vb)}")
    print(f"log p at z_cons: {fit4.log_prob(z_vb[0, :], adjust_transform = False)}")
    # print(f"grad log p at z_cons: {fit4.grad_log_prob(z_vb[0, :], adjust_transform = False)}")
    # print(f"grad log p at z_cons: {fit4.grad_log_prob(z_vb[0, :], adjust_transform = True)}")
    print(f"log p at z_uncons {fit4.log_prob(unconstrain_params(z_vb)[0,:], adjust_transform = True)}")
    print(f"log p at z_uncons {fit4.log_prob((z_vb)[0,:], adjust_transform = True)}")
# test_pystan()
# def test_dicts():
#     from collections import OrderedDict
#     a = {"s": "S", "t" : "T"}
#     b = OrderedDict(**a)
#     print(a)
#     print(a)
#     print(a)
#     print(a)
#     print(a)
#     print("ordered")
#     print(b)
#     print(b)
#     print(b)
#     print(b)
#     print(b)
# test_dicts()
def test_imports():
    import inference
    import interface

test_imports()

"""
We need to take care of verbose.
We need to clean the UI to take only model and data as inputs.
We need to put this all in a nice pip package.
"""


import functools
import autograd.numpy as np
import autograd.scipy.special as autoscipy
import autograd
import vistan.utilities as utils

def objective_utils(params, log_p, log_q, sample_q, \
                                        M_training, num_copies_training):

    samples_shape = (num_copies_training, M_training)  

    params_stopped = autograd.core.getval(params)

    z = sample_q(params, samples_shape)
    
    lp = log_p(z)
    
    lq = log_q(params, z)

    lq_stopped = log_q(params_stopped, z)

    return z, lp, lq, lq_stopped

def ELBO_cf_entropy(params, log_p, log_q, sample_q, entropy_q, \
                                                M_training, num_copies_training):

    _, lp, _, _ = objective_utils(params, log_p, log_q, sample_q, \
                                                M_training, num_copies_training)

    return np.mean(lp) + entropy_q(params)

def IWELBO(params, log_p, log_q, sample_q, M_training, num_copies_training):

    _, lp, lq, _ = objective_utils(params, log_p, log_q, sample_q,\
                                                 M_training, num_copies_training)

    return np.mean(autoscipy.logsumexp(lp - lq, -1)) - np.log(M_training)

def IWELBO_STL(params, log_p, log_q, sample_q, M_training, num_copies_training):

    _, lp, _, lq_stopped = objective_utils(params, log_p, log_q,\
                                     sample_q, M_training, num_copies_training)

    lR = lp - lq_stopped

    return np.mean(np.sum(utils.softmax_matrix(lR)*lR, -1))

def IWELBO_DREG(params, log_p, log_q, sample_q,\
                                 M_training, num_copies_training):

    _, lp, _, lq_stopped = objective_utils(params, log_p,\
                             log_q, sample_q, M_training, num_copies_training)

    lR = lp - lq_stopped

    return np.mean(np.sum((utils.softmax_matrix(lR)**2)*lR, -1))

def choose_objective_eval_fn(hyper_params):

    if hyper_params['grad_estimator'] == "Total-gradient":
        objective = IWELBO 

    elif hyper_params['grad_estimator'] == "DReG":
        objective = IWELBO_DREG  

    elif hyper_params['grad_estimator'] == "STL":
        objective = IWELBO_STL  

    elif hyper_params['grad_estimator'] == "closed-form-entropy":
        assert (hyper_params['M_training'] == 1)
        assert ("gaussian" in hyper_params['vi_family'])
        objective = ELBO_cf_entropy  
    else:
        raise ValueError

    if hyper_params['evaluation_fn'] == "IWELBO":
        evaluation_fn = IWELBO

    elif hyper_params['evaluation_fn'] == "ELBO-cfe":
        evaluation_fn = ELBO_cf_entropy

    else: 
        raise ValueError

    return objective, evaluation_fn

def modify_objective_eval_fn(objective, evaluation_fn,\
                                             log_p, var_dist, hyper_params):

    m_objective = functools.partial(objective, 
                                log_p = log_p,
                                log_q = var_dist.log_prob, 
                                sample_q = var_dist.sample,
                                M_training = hyper_params['M_training'], 
                                num_copies_training = hyper_params['num_copies_training'])

    m_evaluation_fn = functools.partial(evaluation_fn, 
                                log_p = log_p,
                                log_q = var_dist.log_prob, 
                                sample_q = var_dist.sample,
                                M_training = hyper_params['M_training'], 
                                num_copies_training = hyper_params['num_copies_training'])

    if hyper_params['grad_estimator'] == "closed-form-entropy":
        m_objective = functools.partial(m_objective, entropy_q = var_dist.entropy)

    if hyper_params['evaluation_fn'] == "ELBO-cfe":
        m_objective = functools.partial(m_objective, entropy_q = var_dist.entropy)

    # Augment the objective definition to match the Autograd's optimizer template  
    return lambda params,t : -m_objective(params), m_evaluation_fn


def get_objective_eval_fn(log_p, var_dist, hyper_params):

    objective, evaluation_fn = choose_objective_eval_fn(hyper_params)
    
    return  modify_objective_eval_fn(objective, evaluation_fn, log_p, \
                                                        var_dist, hyper_params) 




import functools
import autograd
import autograd.numpy.random as npr

import vistan.vi_families as vi_families
import vistan.objectives as objectives
import vistan.interface as interface
import vistan.utilities as utils
# from utilities.result_helper import (save_results_parameters)


def hyperparams(**kwargs):
    hparams = {
        "seed" : 11,
        
        ##################################################################
        #  Methods 
        ##################################################################
        # choosing a method gives your some default hyper-parameter settings
        # choose from : ['advi, 'gaussian, 'flows', 'meanfield', 'custom']
        # choosing a method will change the values for other relevant 
        # hyper-parameters; however, choosing 'custom', will leave them as is.
        "method" : "meanfield",

        ##################################################################
        #  ADVI hparams
        ##################################################################
        "advi_use": False,
        # The threshold parameters is used to diagnose convergence. If the 
        # mean or median of the relative tolerance window is below the 
        # threshold parameter, then the optimization is stopped.  
        'advi_convergence_threshold' : 0.001, 

        "advi_step_size": 0.01,
        # the range used in the original paper. 
        # Expand this range if the optimization diverges for all these choices
        "advi_adapt_step_size_range": [0.1, 0.01],
        # Whether to adapt the step-size to the problem based on ADVI heuristic or not.
        "advi_adapt_step_size": True,
        # Number of iterations when adapting the step-size. ADVI optimizes each step-size 
        # for these many iterations, looks at the final ELBO, and greedily chooses 
        # the largest step-size with highest final ELBO  
        'advi_adapt_step_size_num_iters': 20, 

        # Number of iterations after which ELBO is calculated. This is coupled with the 
        # size of the circular buffer to detect convergence. During final vi optimization,
        # after each 'advi_callback_iteration', we evaluate ELBO, and append it to a buffer.
        # Then convergence is detected based on the mean or median relative tolerance 
        # over this circular buffer.

        'advi_callback_iteration': 2,

        ##################################################################
        #  Optimization
        ##################################################################
        # step_size = step_size_base/(step_size_scale**step_size_exp)
        # For comprehensive step-search, we optimize for different step_sizes,
        # where the different step_sizes are generated using step_size_exp_range
        "step_size_base" : 0.1,
        "step_size_scale" : 4.0,
        "step_size_exp" : 2, 
        # "step_size_exp_range" : [2], # For comprehensive step_search we suggest [0,1,2,3,4]

        "max_iters" : 20,

        #  Choices for optimizers: "adam", "advi"
        "optimizer":"adam" ,
        #  M_training = 1 corresponds to regular VI -- ELBO is optimized
        #  M_training > 1 corresponds to IW-training
        "M_training"  : 2, 

        #  Four choices for gradient estimator type: "Total-gradient", "STL", "DReG", "closed-form-entropy"
        #  closed-form-entropy will work with distributions like Gaussians.
        #  IWAEDREG defaults to STL when M_training = 1  
        "grad_estimator": "DReG", 

        # Fixing the sample budget. If the M_training = 10, then we make 10 copies of 
        # IW-ELBO with M = 10 at each iteration.
        "per_iter_sample_budget":100, 

        "evaluation_fn" : "IWELBO",

        ##################################################################
        #  VI Families
        ##################################################################

        #  Choices for vi_vi_families: "rnvp", "gaussian", "diagonal"
        "vi_family" : "gaussian",

        ##################################################################
        #  RealNVP
        ##################################################################

        # of transformations : # of coupling transforms; each transforms includes two transitions
        # of hidden layers : # of hidden layers within s-t block
        # of hidden units : # of units in each hidden layer
        # params_init_scale : lower value initializes to standard normal approximately

        "rnvp_num_transformations" : 10,
        "rnvp_num_hidden_units" : 32,
        "rnvp_num_hidden_layers" : 2,
        "rnvp_params_init_scale" : 0.01,


        ##################################################################
        #  LI
        ##################################################################
        #  Do not use LI with real-NVP
        "LI" : False,
        "LI_max_iters":2000,
        "LI_epsilon":1e-6,

    }
    
    for k in kwargs.keys():
        if k not in hparams.keys():
            raise KeyError(f"{k} is not an expected hyperparam.")
    hparams.update(kwargs)

    return hparams


def inference(code, data, model_name = "test-model", verbose = True, hparams = hyperparams()):
    npr.seed(hparams['seed'])


    # Stan is particular about model names
    if model_name is not None: 
        model_name = model_name.replace("-", "_")

    model = interface.Model(code, data, model_name, verbose = verbose)

    utils.update_hparams(model, hparams)
    print("Printing the Hyper-param configuration file...")
    utils.print_dict(hparams)

    if hparams['advi_use'] == True:
        utils.advi_asserts(hparams)

    var_dist = vi_families.get_var_dist(hparams)
    init_params = var_dist.initial_params()

    if hparams['LI'] == 1:
        # update the initial parameters to LI params
        assert "gaussian" in hparams['vi_family'] 
        init_params = utils.get_laplaces_init (log_p = model.log_prob, \
                                    z_len = hparams['latent_dim'], \
                                    num_epochs = hparams['LI_max_iters'], \
                                    ε = hparams['LI_epsilon'],\
                                    model_name = model_name,
                                    model_code = code)

    log_p = model.log_prob
    log_q = var_dist.log_prob
    sample_q = var_dist.sample

    objective, eval_function = objectives.get_objective_eval_fn(log_p, var_dist,\
                                                                    hparams)
    objective_grad = autograd.grad(objective) 

    optimizer = utils.get_optimizer(hparams)
    callback = utils.get_callback(hparams)
    print("Starting optimization......")
    results, optimized_params = utils.optimization_handler(objective_grad = objective_grad,
                                                    eval_function = eval_function,
                                                    init_params = init_params,
                                                    optimizer = optimizer,
                                                    num_epochs = hparams['max_iters'],
                                                    step_size = utils.get_step_size(hparams),
                                                    callback = functools.partial(callback,
                                                                    model = model, 
                                                                    eval_function = eval_function),
                                                    verbose = verbose,
                                                    hparams = hparams)
    posterior = vi_families.get_posterior(model = model, var_dist = var_dist,\
                                                M_sampling = hparams['M_training'], params = optimized_params)

    return posterior, model, results